/*         ______   ___    ___
 *        /\  _  \ /\_ \  /\_ \
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *
 *      ALSA 0.9 sound driver.
 *
 *      By Thomas Fjellstrom.
 *
 *      Extensively modified by Elias Pschernig.
 *
 *      Rewritten for 4.3 sound API by Milan Mimica.
 *
 *      See readme.txt for copyright information.
 */


#include "internal.h"

#ifndef SCAN_DEPEND
   #include <string.h>
   #define ALSA_PCM_NEW_HW_PARAMS_API 1
   #include <alsa/asoundlib.h>
   #include <math.h>
#endif



/* To be put into ALSA9_CHECK macro. */
/*      uszprintf(allegro_error, ALLEGRO_ERROR_SIZE, "ALSA: %s : %s",
               #a, get_config_text(snd_strerror(err)));*/

#define ALSA9_CHECK(a) do {                                           \
   int err = (a);                                                     \
   if (err < 0) {                                                     \
      fprintf(stderr, "%s: %s\n", strerror(-err), #a);                \
      goto Error;                                                     \
   }                                                                  \
} while(0)


#define GET_CHANNEL_COUNT(v) ((v>>4)+(v&0xF))


static int alsa_channel_count_pref = 2;
static int alsa_depth_pref = 16;
static unsigned int alsa_freq_pref = 44100;
static unsigned int alsa_frag_count_pref = 5;
static snd_pcm_uframes_t alsa_frag_size_pref = 256;
static char alsa_device_pref[128] = "default";
static char alsa_mixer_device_pref[128] = "default";

static int alsa_channel_count;
static int alsa_depth;
static unsigned int alsa_freq;
static unsigned int alsa_frag_count;
static snd_pcm_uframes_t alsa_frag_size;
static char alsa_mixer_device[128];

static snd_pcm_t *alsa_pcm_handle = NULL;
static snd_mixer_t *alsa_mixer = NULL;

static snd_output_t *snd_output = NULL;

static pthread_t alsa_poll_thread = 0;
static bool alsa_quit_poll_thread;


typedef struct ALSA_VOICE {
   unsigned char *buffer;
   unsigned int buf_len;
   unsigned int pos;

   volatile bool stopped;
   AL_AUDIO_ENUM loop_mode;
} ALSA_VOICE;

static ALSA_VOICE *alsa_voice = NULL;
_AL_MUTEX alsa_voice_mutex;



/* The open method starts up the driver and should lock the device, using the
   previously set paramters, or defaults. It shouldn't need to start sending
   audio data to the device yet, however. */
static int alsa_open()
{
   int ret;
   snd_pcm_uframes_t req_frag_size = alsa_frag_size_pref;
   snd_pcm_format_t format;

   snd_pcm_hw_params_t *hwparams;
   snd_pcm_sw_params_t *swparams;

   alsa_channel_count = alsa_channel_count_pref;
   alsa_depth = alsa_depth_pref;
   strcpy(alsa_mixer_device, alsa_mixer_device_pref);

   ALSA9_CHECK(snd_output_stdio_attach(&snd_output, stdout, 0));

   ret = snd_pcm_open(&alsa_pcm_handle, alsa_device_pref, SND_PCM_STREAM_PLAYBACK, 0);
   if (ret < 0) {
      //TODO: enable this
      //ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Can not open card/pcm device"));
      return -1;
   }

   ALSA9_CHECK(snd_mixer_open(&alsa_mixer, 0));

   if (alsa_mixer
      && snd_mixer_attach(alsa_mixer, alsa_mixer_device) >= 0
      && snd_mixer_selem_register (alsa_mixer, NULL, NULL) >= 0
      && snd_mixer_load(alsa_mixer) >= 0) {
      const char *name;

      snd_mixer_elem_t *mixer_elem = snd_mixer_first_elem(alsa_mixer);

      do {
         name = snd_mixer_selem_get_name(mixer_elem);
         mixer_elem = snd_mixer_elem_next(mixer_elem);
      } while (mixer_elem && strcasecmp(name, alsa_mixer_device));

      if (mixer_elem) {
         snd_mixer_selem_set_playback_volume_range(mixer_elem, 0, 255);
      }
   }

   if (alsa_depth == 8)
      format = SND_PCM_FORMAT_U8;
   else if (alsa_depth == 16)
      format = SND_PCM_FORMAT_S16;
   else if (alsa_depth == 24)
      format = SND_PCM_FORMAT_S24;
   else
      format = SND_PCM_FORMAT_S16;

   snd_pcm_hw_params_alloca(&hwparams);
   snd_pcm_sw_params_alloca(&swparams);

   ALSA9_CHECK(snd_pcm_hw_params_any(alsa_pcm_handle, hwparams));
   ALSA9_CHECK(snd_pcm_hw_params_set_access(alsa_pcm_handle, hwparams, SND_PCM_ACCESS_RW_INTERLEAVED));
   ALSA9_CHECK(snd_pcm_hw_params_set_format(alsa_pcm_handle, hwparams, format));
   ALSA9_CHECK(snd_pcm_hw_params_set_channels(alsa_pcm_handle, hwparams, alsa_channel_count));

   ALSA9_CHECK(snd_pcm_hw_params_set_rate_near(alsa_pcm_handle, hwparams, &alsa_freq_pref, NULL));
   ALSA9_CHECK(snd_pcm_hw_params_set_period_size_near(alsa_pcm_handle, hwparams, &alsa_frag_size_pref, NULL));
   ALSA9_CHECK(snd_pcm_hw_params_set_periods_near(alsa_pcm_handle, hwparams, &alsa_frag_count_pref, NULL));

   ALSA9_CHECK(snd_pcm_hw_params(alsa_pcm_handle, hwparams));

   ALSA9_CHECK(snd_pcm_hw_params_get_rate(hwparams, &alsa_freq, NULL));
   ALSA9_CHECK(snd_pcm_hw_params_get_period_size(hwparams, &alsa_frag_size, NULL));
   ALSA9_CHECK(snd_pcm_hw_params_get_periods(hwparams, &alsa_frag_count, NULL));


   ALSA9_CHECK(snd_pcm_sw_params_current(alsa_pcm_handle, swparams));
   ALSA9_CHECK(snd_pcm_sw_params_set_start_threshold(alsa_pcm_handle, swparams, alsa_frag_size));
   ALSA9_CHECK(snd_pcm_sw_params_set_avail_min(alsa_pcm_handle, swparams, req_frag_size));
   ALSA9_CHECK(snd_pcm_sw_params_set_xfer_align(alsa_pcm_handle, swparams, 1));
   ALSA9_CHECK(snd_pcm_sw_params(alsa_pcm_handle, swparams));

   /*uszprintf(alsa_desc, sizeof(alsa_desc),
         get_config_text
         ("Alsa 0.9, device '%s': %d bits (format %s), %d bps, %d channels\n"),
         alsa_device, alsa_depth, snd_pcm_format_name(format),
         alsa_freq, alsa_channel_count);*/

   printf("ALSA, device '%s': %d bits (format %s), %d bps, %d channels\n",
          alsa_device_pref, alsa_depth, snd_pcm_format_name(format), alsa_freq,
          alsa_channel_count);

   return 0;

   Error:
   if (alsa_pcm_handle) {
      snd_pcm_close(alsa_pcm_handle);
      alsa_pcm_handle = NULL;
   }

   printf("Error!\n");

   return -1;
}



/* The close method should close the device, freeing any resources, and allow
   other processes to use the device */
static void alsa_close()
{
   if (alsa_mixer) {
      snd_mixer_detach(alsa_mixer, alsa_mixer_device);
      snd_mixer_free(alsa_mixer);
      snd_mixer_close(alsa_mixer);
   }

   if (alsa_pcm_handle)
      snd_pcm_close(alsa_pcm_handle);

   snd_config_update_free_global();
}



/* Inverts the content of the voice buffer so that it is played reversed */
static void alsa_invert_voice_buffer()
{
   if (!alsa_voice || !alsa_voice->buffer)
      return;

   unsigned int x;
   const int w = alsa_depth / 8 * alsa_channel_count;
   unsigned char *buf;

   buf = malloc(alsa_voice->buf_len);

   for (x = 0; x < alsa_voice->buf_len / w; ++x) {
      memcpy(buf + x*w, alsa_voice->buffer + alsa_voice->buf_len - x*w - w, w);
   }

   free(alsa_voice->buffer);
   alsa_voice->buffer = buf;
}



/* Underrun and suspend recovery */
static int xrun_recovery(int err)
{
   if (err == -EPIPE) { /* under-run */
      err = snd_pcm_prepare(alsa_pcm_handle);
      if (err < 0) {
         fprintf(stderr, "Can't recovery from underrun, prepare failed: %s\n", snd_strerror(err));
      }
      return 0;
   }
   else if (err == -ESTRPIPE) {
      while ((err = snd_pcm_resume(alsa_pcm_handle)) == -EAGAIN) {
         //TODO: replace with al_rest
         usleep(10000); /* wait until the suspend flag is released */
      }
      if (err < 0) {
         err = snd_pcm_prepare(alsa_pcm_handle);
         if (err < 0) {
            fprintf(stderr, "Can't recovery from suspend, prepare failed: %s\n", snd_strerror(err));
         }
      }
      return 0;
   }
   return err;
}



/* Custom routine which runs in another thread and fills the harware PCM buffer with data
   from the stream. */
static int alsa_update_stream(AL_VOICE *voice, void **buf)
{
   int buf_len, frame_len;

   frame_len = (alsa_depth / 8) * alsa_channel_count;
   buf_len = alsa_frag_size;
   *buf = (unsigned char *)_al_voice_update(voice, buf_len / frame_len);
   if (!*buf) {
      return 0;
   }

   return buf_len;
}


/* Custom routine which runs in another thread and fills the hardware PCM buffer
   from the voice buffer. */
static int alsa_update_nonstream(void **buf)
{
   int buf_len, pos;

   if (alsa_voice->pos == alsa_voice->buf_len) {
      alsa_voice->pos = 0;
      if (alsa_voice->loop_mode == AL_AUDIO_PLAY_ONCE) {
         alsa_voice->stopped = true;
         return 0;
      }
      else if (alsa_voice->loop_mode == AL_AUDIO_BI_DIR) {
         alsa_invert_voice_buffer();
      }
   }

   pos = alsa_voice->pos;

   if (alsa_voice->pos + alsa_frag_size < alsa_voice->buf_len) {
      alsa_voice->pos += alsa_frag_size;
   }
   else {
      alsa_voice->pos = alsa_voice->buf_len;
   }

   buf_len = alsa_voice->pos - pos;
   *buf = alsa_voice->buffer + pos;

   return buf_len;
}


static void *alsa_update(void *arg)
{
   AL_VOICE *voice = (AL_VOICE*)arg;
   void *buf, *tmp_buf;
   int res;

   tmp_buf = malloc(alsa_frag_size);

   while (!alsa_quit_poll_thread) {
      _al_mutex_lock(&alsa_voice_mutex);

      res = 0;
      if (!alsa_voice->stopped) {
         if (!voice->streaming)
            res = alsa_update_nonstream(&buf);
         else
            res = alsa_update_stream(voice, &buf);
      }

      /* We need to pass a copy of buffer fragment to alsa because
         snd_pcm_writei() is blocking and we have to unlock the mutex before
         it in order to have it unlocked for a while. */
      if (!res) {
         res = alsa_frag_size;
         if (alsa_depth == 8)
            memset(tmp_buf, 0x80, res);
         else
            memset(tmp_buf, 0, res);
      }
      else
         memcpy(tmp_buf, buf, res);

      _al_mutex_unlock(&alsa_voice_mutex);

      res = snd_pcm_writei(alsa_pcm_handle, tmp_buf, res / (alsa_depth / 8) / alsa_channel_count);
      if (res < 0) {
         if (xrun_recovery(res) < 0) {
            printf("Write error: %s\n", snd_strerror(res));
            continue;
         }
      }
   }

   free(tmp_buf);
   return NULL;
}



/* The load_voice method loads a sample into the driver's memory. The voice's
   'streaming' field will be set to false for these voices, and it's
   'buffer_size' field will be the total length in bytes of the sample data.
   The voice's attached stream's looping mode should be honored, and loading
   must fail if it cannot be. */
static int alsa_load_voice(AL_VOICE *voice, const void *data)
{
   ALSA_VOICE *ex_data = voice->extra;

   _al_mutex_lock(&alsa_voice_mutex);

   ex_data->buffer = malloc(voice->buffer_size);
   memcpy(ex_data->buffer, data, voice->buffer_size);

   ex_data->buf_len = voice->buffer_size;
   if(!ex_data->buf_len)
      ex_data->buf_len = alsa_frag_size;

   ex_data->pos = 0;
   ex_data->stopped = true;
   ex_data->loop_mode = voice->stream->loop;

   _al_mutex_unlock(&alsa_voice_mutex);

   return 0;
}



/* The unload_voice method unloads a sample previously loaded with load_voice.
   This method should not be called on a streaming voice. */
static void alsa_unload_voice(AL_VOICE *voice)
{
   ALSA_VOICE *ex_data = voice->extra;

   _al_mutex_lock(&alsa_voice_mutex);

   free(ex_data->buffer);
   memset(ex_data, 0, sizeof(struct ALSA_VOICE));
   ex_data->stopped = true;

   _al_mutex_unlock(&alsa_voice_mutex);
}



/* The start_voice should, surprise, start the voice. For streaming voices, it
   should start polling the device and call _al_voice_update for audio data.
   For non-streaming voices, it should resume playing from the last set
   position */
static int alsa_start_voice(AL_VOICE *voice)
{
   ALSA_VOICE *ex_data = voice->extra;

   _al_mutex_lock(&alsa_voice_mutex);
   ex_data->stopped = false;
   _al_mutex_unlock(&alsa_voice_mutex);

   return 0;
}



/* The stop_voice method should stop playback. For non-streaming voices, it
   should leave the data loaded, and reset the voice position to 0. */
static void alsa_stop_voice(AL_VOICE *voice)
{
   ALSA_VOICE *ex_data = voice->extra;

   _al_mutex_lock(&alsa_voice_mutex);
   ex_data->stopped = true;
   if(!voice->streaming) {
      ex_data->pos = 0;
   }
   _al_mutex_unlock(&alsa_voice_mutex);
}



/* The voice_is_playing method should only be called on non-streaming sources,
   and should return true if the voice is playing */
static bool alsa_voice_is_playing(const AL_VOICE *voice)
{
   ALSA_VOICE *ex_data = voice->extra;
   return !ex_data->stopped;
}



/* The allocate_voice method should grab a voice from the system, and allocate
   any data common to streaming and non-streaming sources. */
static int alsa_allocate_voice(AL_VOICE *voice)
{
   /* For now, we support only one hardware voice. */
   if (alsa_voice)
      return -1;

   alsa_voice = calloc(1, sizeof(ALSA_VOICE));
   if (!alsa_voice)
      return -1;
   alsa_voice->stopped = true;

   _al_mutex_init(&alsa_voice_mutex);
   alsa_quit_poll_thread = false;
   if (pthread_create(&alsa_poll_thread, NULL, alsa_update, voice))
      return -1;

   voice->extra = alsa_voice;

   return 0;
}



/* The deallocate_voice method should free the resources for the given voice,
   but still retain a hold on the device. The voice should be stopped and
   unloaded by the time this is called */
static void alsa_deallocate_voice(AL_VOICE *voice)
{
   if (!alsa_voice)
      return;

   alsa_quit_poll_thread = true;
   pthread_join(alsa_poll_thread, NULL);

   free(voice->extra);
   voice->extra = NULL;
   alsa_voice = NULL;

   _al_mutex_deinit(&alsa_voice_mutex);
}



/* The get_voice_position method should return the current sample position of
   the voice (sample_pos = byte_pos / (depth/8) / channels). This should never
   be called on a streaming voice. */
static unsigned long alsa_get_voice_position(const AL_VOICE *voice)
{
   ALSA_VOICE *ex_data = voice->extra;
   return ex_data->pos * 8 / (voice->depth * GET_CHANNEL_COUNT(voice->chan_conf));
}



/* The set_voice_position method should set the voice's playback position,
   given the value in samples. This should never be called on a streaming
   voice. */
static int alsa_set_voice_position(AL_VOICE *voice, unsigned long val)
{
   ALSA_VOICE *ex_data = voice->extra;
   if (ex_data->buf_len < val)
      return -1;
   else {
      _al_mutex_lock(&alsa_voice_mutex);
      ex_data->pos = val * GET_CHANNEL_COUNT(voice->chan_conf) * voice->depth / 8;
      _al_mutex_unlock(&alsa_voice_mutex);
   }

   return 0;
}



/* A good portion of the get/set methods will use driver-specific option
   values, though it is strongly encouraged to support a standard set of
   options. The set methods may be called without the driver being initialized,
   and generally shouldn't take effect until the next time the driver is
   initialized. The get methods are allowed to fail (cleanly) if the driver
   isn't currently in use and should return the values being used (not
   necesarilly what was previously set) */

static int alsa_set_bool(AL_AUDIO_ENUM setting, bool val)
{
   return 1;
   (void)setting;
   (void)val;
}



static int alsa_set_enum(AL_AUDIO_ENUM setting, AL_AUDIO_ENUM val)
{
   switch(setting) {

      case AL_AUDIO_DEPTH:
         if (val == AL_AUDIO_8_BIT_INT)
            alsa_depth_pref = 8;
         else if (val == AL_AUDIO_16_BIT_INT)
            alsa_depth_pref = 16;
         else if (val == AL_AUDIO_24_BIT_INT)
            alsa_depth_pref = 24;
         else
            return 1;
      return 0;

      case AL_AUDIO_CHANNELS:
         alsa_channel_count_pref = GET_CHANNEL_COUNT(val);
      return 0;

      default:
         break;
   }

   return 1;
}



/* With the set_long method, the following options would be useful:
 * * AL_AUDIO_FREQUENCY - the over-all output frequency (in hz)
 * * AL_AUDIO_FRAGMENTS - total fragment count (# of buffers for the output
 *                        device)
 * * AL_AUDIO_LENGTH    - the size of each fragment (in bytes)
 * The values set are merely hints and don't need to be followed precisely.
 */
static int alsa_set_long(AL_AUDIO_ENUM setting, unsigned long val)
{
   switch(setting) {

      case AL_AUDIO_FREQUENCY:
         if (val)
            alsa_freq_pref = val;
      return 0;

      case AL_AUDIO_FRAGMENTS:
         if(val)
            alsa_frag_count_pref = val;
      return 0;

      case AL_AUDIO_LENGTH:
         if(val)
            alsa_frag_size_pref = val;
      return 0;

      default:
         break;
   }

   return 1;
}



/* The set_ptr method should be able to take the following:
 * * AL_AUDIO_DEVICE - The device the driver should use, represented as a
 *                     string. The pointer should not be assumed to remain
 *                     valid after the call is completed.
 */
static int alsa_set_ptr(AL_AUDIO_ENUM setting, const void *val)
{
   switch(setting) {

      case AL_AUDIO_DEVICE:
         if(val)
            strncpy(alsa_device_pref, val, sizeof(alsa_device_pref));
         else
            alsa_device_pref[0] = '\0';
      return 0;

      case AL_ALSA_MIXER_DEVICE:
         if (val)
            strncpy(alsa_mixer_device_pref, val, sizeof(alsa_mixer_device_pref));
         else
            alsa_mixer_device_pref[0] = '\0';
      return 0;

      default:
         break;
   }

   return 1;
}



static int alsa_get_bool(AL_AUDIO_ENUM setting, bool *val)
{
   return 1;
   (void)setting;
   (void)val;
}



static int alsa_get_enum(AL_AUDIO_ENUM setting, AL_AUDIO_ENUM *val)
{
   snd_pcm_hw_params_t *hwparams;

   if (!alsa_pcm_handle)
      return 1;

   snd_pcm_hw_params_alloca(&hwparams);
   snd_pcm_hw_params_current(alsa_pcm_handle, hwparams);

   switch(setting) {

      case AL_AUDIO_DEPTH:
      {
         snd_pcm_format_t format;
         if (snd_pcm_hw_params_get_format(hwparams, &format) < 0)
            return 1;
         else {
            if (format == SND_PCM_FORMAT_S8)
               *val = AL_AUDIO_8_BIT_INT;
            else if (format == SND_PCM_FORMAT_U8)
               *val = AL_AUDIO_8_BIT_UINT;
            else if (format == SND_PCM_FORMAT_S16_LE || format == SND_PCM_FORMAT_S16_BE || SND_PCM_FORMAT_S16)
               *val = AL_AUDIO_16_BIT_INT;
            else if (format == SND_PCM_FORMAT_U16_LE || format == SND_PCM_FORMAT_U16_BE || SND_PCM_FORMAT_U16)
               *val = AL_AUDIO_16_BIT_UINT;
            else if (format == SND_PCM_FORMAT_S24_LE || format == SND_PCM_FORMAT_S24_BE || SND_PCM_FORMAT_S24)
               *val = AL_AUDIO_24_BIT_INT;
            else if (format == SND_PCM_FORMAT_U24_LE || format == SND_PCM_FORMAT_U24_BE || SND_PCM_FORMAT_U24)
               *val = AL_AUDIO_24_BIT_UINT;
            else
               return 1;

            return 0;
         }
      }

      case AL_AUDIO_CHANNELS:
      {
         unsigned int channels;
         if (snd_pcm_hw_params_get_channels(hwparams, &channels) < 0)
            return 1;
         else {
            if (channels == 1)
               *val = AL_AUDIO_1_CH;
            else if (channels == 2)
               *val = AL_AUDIO_2_CH;
            else if (channels == 3)
               *val = AL_AUDIO_3_CH;
            else if (channels == 4)
               *val = AL_AUDIO_4_CH;
            else if (channels == 5)
               *val = AL_AUDIO_5_CH;
            else if (channels == 6)
               *val = AL_AUDIO_6_CH;
            else if (channels == 7)
               *val = AL_AUDIO_7_CH;
            else
               return 1;

            return 0;
         }
      }

      default:
         break;
   }

   return 1;
}



/* The get_long method would  be useful to retrieve the following options:
 * * AL_AUDIO_FREQUENCY - The actual output frequency in use by the current
 *                        device
 * * AL_AUDIO_LENGTH    - The length in bytes of each transfer buffer/fragment
 * * AL_AUDIO_FRAGMENTS - The number of fragments
 */
static int alsa_get_long(AL_AUDIO_ENUM setting, unsigned long *val)
{
   snd_pcm_hw_params_t *hwparams;

   if (!alsa_pcm_handle)
      return 1;

   snd_pcm_hw_params_alloca(&hwparams);
   snd_pcm_hw_params_current(alsa_pcm_handle, hwparams);

   switch(setting) {

      case AL_AUDIO_FREQUENCY:
      {
         unsigned int rate;
         if (snd_pcm_hw_params_get_rate(hwparams, &rate, NULL) < 0)
            return 1;
         else {
            *val = (unsigned long)rate;
            return 0;
         }
      }

      case AL_AUDIO_LENGTH:
      {
         if (snd_pcm_hw_params_get_period_size(hwparams, val, NULL) < 0)
            return 1;
         else
            return 0;
      }

      case AL_AUDIO_FRAGMENTS:
      {
         unsigned int periods;
         if (snd_pcm_hw_params_get_periods(hwparams, &periods, NULL) < 0)
            return 1;
         else {
            *val = (unsigned long)periods;
            return 0;
         }
      }

      default:
         break;
   }

   return 1;
}



/* The get_ptr method should handle the following:
 * * AL_AUDIO_DEVICE - The name of the current device, given as a string. The
 *                     returned string will not be modified.
 */
static int alsa_get_ptr(AL_AUDIO_ENUM setting, const void **val)
{
   switch(setting) {

      case AL_AUDIO_DEVICE:
         if(!alsa_pcm_handle)
            return 1;
         *val = snd_pcm_name(alsa_pcm_handle);
      return 0;

      case AL_ALSA_MIXER_DEVICE:
         if (!alsa_pcm_handle)
            return 1;
         *val = alsa_mixer_device;
      return 0;

      default:
         break;
   }

   return 1;
}



AL_AUDIO_DRIVER _alsa_driver =
{
   "ALSA",

   alsa_open,
   alsa_close,

   alsa_get_bool,
   alsa_get_enum,
   alsa_get_long,
   alsa_get_ptr,

   alsa_set_bool,
   alsa_set_enum,
   alsa_set_long,
   alsa_set_ptr,

   alsa_allocate_voice,
   alsa_deallocate_voice,

   alsa_load_voice,
   alsa_unload_voice,

   alsa_start_voice,
   alsa_stop_voice,

   alsa_voice_is_playing,

   alsa_get_voice_position,
   alsa_set_voice_position,
};
