#include "allegro.h"

#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>


void display_driver_info(AL_AUDIO_DRIVER *driver);
void display_voice_info(AL_VOICE *voice);
void display_mixer_info(AL_MIXER *mixer);
void stream_file(const char *filename);

/* Function that allegro usually provides. */
AL_SAMPLE* allegro_load_sample(const char *filename);

AL_AUDIO_ENUM chan, depth;
unsigned long freq;
AL_AUDIO_DRIVER *driver;


int main(int argc, char **argv) {
	int x;
	bool stream_it = false;

	if(argc < 2) {
		fprintf(stderr, "Usage: %s [-stream] <sample.wav>...\n", argv[0]);
		return 1;
	}

	if(strcasecmp(argv[1], "-stream") == 0)
	{
		++argv;
		--argc;
		stream_it = true;
	}

	allegro_init();

    al_error_set_handler(NULL);

	driver = al_get_audio_driver("alsa");
//	driver = al_get_audio_driver("openal");
	if(driver) {
		al_audio_set_enum(driver, AL_AUDIO_DEPTH, AL_AUDIO_16_BIT_INT);
		al_audio_set_enum(driver, AL_AUDIO_CHANNELS, AL_AUDIO_2_CH);
		al_audio_set_long(driver, AL_AUDIO_FREQUENCY, 44100);
	}
	driver = al_audio_init_driver(driver);
	if (!driver) {
		fprintf(stderr, "Could not init sound!\n");
		return 1;
	}

	if(al_audio_get_enum(driver, AL_AUDIO_DEPTH, &depth))
		depth = AL_AUDIO_16_BIT_INT;
	if(al_audio_get_enum(driver, AL_AUDIO_CHANNELS, &chan))
		chan = AL_AUDIO_2_CH;
	if(al_audio_get_long(driver, AL_AUDIO_FREQUENCY, &freq))
		freq = 44100;

	display_driver_info(driver);

	for (x = 1; x < argc; ++x) {
		AL_SAMPLE *sample = NULL;
		AL_VOICE *voice = NULL;
		AL_MIXER *mixer = NULL;
		AL_AUDIO_ENUM chan_conf;
		unsigned long voice_freq;

		const char *filename = argv[x];

		if (stream_it) {
			stream_file(filename);
			continue;
		}

		sample = allegro_load_sample(filename);
		if (!sample) {
			fprintf(stderr, "Could not create AL_SAMPLE from '%s'!\n", filename);
			continue;
		}

		//al_sample_set_enum(sample, AL_AUDIO_LOOPMODE, AL_AUDIO_BI_DIR);

		voice = al_voice_create(driver, freq, depth, chan, 0);
		if(!voice) {
			fprintf(stderr, "Error allocating voice!\n");
			goto getout;
		}
		display_voice_info(voice);
		al_voice_get_long(voice, AL_AUDIO_FREQUENCY, &voice_freq);
		al_voice_get_enum(voice, AL_AUDIO_CHANNELS, &chan_conf);

		mixer = al_mixer_create(voice_freq, AL_AUDIO_32_BIT_FLOAT, chan_conf);
		if(!mixer) {
			fprintf(stderr, "Error allocating mixer!\n");
			goto getout;
		}

		if(al_voice_attach_mixer(voice, mixer) != 0)
		{
			fprintf(stderr, "Error attaching mixer to voice!\n");
			goto getout;
		}

		/* Attach the sample to the mixer */
		if(al_mixer_attach_sample(mixer, sample) != 0)
		{
			fprintf(stderr, "Error attaching sample to mixer!\n");
			goto getout;
		}

		display_mixer_info(mixer);

		/* Attach the sample to the voice (frequency. channel configuration, and
		* bit depth/signedness must match) */
		/*if(al_voice_attach_sample(voice, sample) != 0)
		{
			fprintf(stderr, "Error attaching sample to voice!\n");
			goto getout;
		}*/

		fprintf(stderr, "Playing '%s'...", filename);

		al_sample_play(sample);
		{
			bool playing;
			do {
				usleep(1000);
				al_sample_get_bool(sample, AL_AUDIO_PLAYING, &playing);
			}while(playing);

			/* The sample may be marked inactive before it is fully played by the
			card. Waiting an extra second will make sure we hear it all */
			usleep(999999);
		}
		fprintf(stderr, "\n");

		al_sample_set_bool(sample, AL_AUDIO_ATTACHED, 0);
		al_mixer_set_bool(mixer, AL_AUDIO_ATTACHED, 0);

getout:
		al_mixer_destroy(mixer);
		al_voice_destroy(voice);
		al_sample_destroy(sample);
	}

	al_audio_deinit_driver(driver);

	return 0;
} END_OF_MAIN()


void display_driver_info(AL_AUDIO_DRIVER *driver) {
	const void *devname;
	unsigned long freq;
	AL_AUDIO_ENUM val;

	if(al_audio_get_ptr(driver, AL_AUDIO_DEVICE, &devname) == 0)
		fprintf(stderr, "Output driver: %s\n", (const char*)devname);
	else
		fprintf(stderr, "Could not get driver name!\n");

	if(al_audio_get_long(driver, AL_AUDIO_FREQUENCY, &freq) == 0)
		fprintf(stderr, "Output frequency: %lu\n", freq);
	else
		fprintf(stderr, "Could not get output frequency!\n");

	if(al_audio_get_enum(driver, AL_AUDIO_CHANNELS, &val) == 0)
		fprintf(stderr, "Output on %d.%d channels.\n", val>>4, val&0xF);
	else
		fprintf(stderr, "Could not get channel config!\n");
	fprintf(stderr, "\n");
}


void display_voice_info(AL_VOICE *voice) {
	unsigned long voice_freq;
	AL_AUDIO_ENUM chan_conf;
	AL_AUDIO_ENUM depth;

	al_voice_get_long(voice, AL_AUDIO_FREQUENCY, &voice_freq);
	al_voice_get_enum(voice, AL_AUDIO_CHANNELS, &chan_conf);
	al_voice_get_enum(voice, AL_AUDIO_DEPTH, &depth);

	fprintf(stderr, "Got voice: %s, %lu hz, %d.%d channel\n",
	        ((depth == AL_AUDIO_8_BIT_INT) ? "8-bit signed" :
	         ((depth == AL_AUDIO_8_BIT_UINT) ? "8-bit unsigned" :
	          ((depth == AL_AUDIO_16_BIT_INT) ? "16-bit signed" :
	           ((depth == AL_AUDIO_16_BIT_UINT) ? "16-bit unsigned" :
	            ((depth == AL_AUDIO_24_BIT_INT) ? "24-bit signed" :
	             ((depth == AL_AUDIO_24_BIT_UINT) ? "24-bit unsigned" :
	              "unknown depth")))))),
	        voice_freq, chan_conf>>4, chan_conf&0xF);
}


void display_mixer_info(AL_MIXER *mixer) {
	AL_AUDIO_ENUM val;

	if(al_mixer_get_enum(mixer, AL_AUDIO_QUALITY, &val) != 0)
		fprintf(stderr, "al_mixer_get_enum failed!\n");
	else
		fprintf(stderr, "%s resampling\n",
	           ((val==AL_AUDIO_POINT) ? "Point" :
	           ((val==AL_AUDIO_LINEAR) ? "Linear" : "Unknown")));
}

struct WAVFILE {
	char *buffer;
	int len;
	int freq;
	int bits;
	int channels;
};


void stream_file(const char *filename) {
	FILE *f;
	struct WAVFILE *wf;

	f = fopen(filename, "rb");
	if(!f)
	{
		fprintf(stderr, "Could not open %s\n", filename);
		return;
	}

	wf = malloc(sizeof(struct WAVFILE));
	wf->len = 0;
	wf->freq = 0;
	wf->bits = 0;
	wf->channels = 0;
	wf->buffer = NULL;

	/* check RIFF header */
	if(fgetc(f)!='R' || fgetc(f)!='I' || fgetc(f)!='F' || fgetc(f)!='F')
	{
		fprintf(stderr, "Bad RIFF header!\n");
		free(wf);
		return;
	}
	fgetc(f); fgetc(f);
	fgetc(f); fgetc(f);
//	fseek(f, 4, SEEK_CUR);

	if(fgetc(f)!='W' || fgetc(f)!='A' || fgetc(f)!='V' || fgetc(f)!='E')
	{
		fprintf(stderr, "Bad WAVE header!\n");
		free(wf);
		return;
	}

	while(1)
	{
		char tag[4];
		int length;

		if(fread(tag, 1, 4, f) != 4)
		{
			fprintf(stderr, "EOF looking for wave data!\n");
			free(wf);
			return;
		}

		/* read chunk length */
		length  = fgetc(f);
		length |= fgetc(f)<<8;
		length |= fgetc(f)<<16;
		length |= fgetc(f)<<24;

		if(memcmp(tag, "fmt ", 4) == 0)
		{
			short i = fgetc(f);
			i |= fgetc(f)<<8;
			length -= 2;
			/* should be 1 for PCM data */
			if (i != 1)
			{
				fprintf(stderr, "Not PCM data!\n");
				free(wf);
				return;
			}

			wf->channels  = fgetc(f);
			wf->channels |= fgetc(f)<<8;
			length -= 2;

			/* sample frequency */
			wf->freq  = fgetc(f);
			wf->freq |= fgetc(f)<<8;
			wf->freq |= fgetc(f)<<16;
			wf->freq |= fgetc(f)<<24;
			length -= 4;

			/* skip six bytes */
			fgetc(f); fgetc(f);
			fgetc(f); fgetc(f);
			fgetc(f); fgetc(f);
			length -= 6;

			wf->bits  = fgetc(f);
			wf->bits |= fgetc(f)<<8;
			length -= 2;

			fprintf(stderr, "Opened file: %d bits, %d hz, %d channel\n",
			        wf->bits, wf->freq, wf->channels);
		}
		else if (memcmp(tag, "data", 4) == 0)
		{
			AL_VOICE *voice;
			AL_MIXER *mixer;

			if(wf->channels == 0 || wf->bits == 0)
				return;

			size_t bpu = 2048 * wf->bits / 8 * ((wf->channels>>4) + (wf->channels&0xF));
			AL_STREAM *stream = al_stream_create(4, 2048, wf->freq,
		                        (wf->bits==8) ? AL_AUDIO_8_BIT_UINT:AL_AUDIO_16_BIT_INT,
		                        (wf->channels==1) ? AL_AUDIO_1_CH:AL_AUDIO_2_CH);
			if(!stream) {
				fprintf(stderr, "Error allocating voice!\n");
				return;
			}

			voice = al_voice_create(driver, freq, (depth==8) ? AL_AUDIO_8_BIT_UINT:AL_AUDIO_16_BIT_INT, chan, 0);
			if(!voice) {
				al_stream_destroy(stream);
				fprintf(stderr, "Error allocating voice!\n");
				return;
			}

			display_voice_info(voice);

			mixer = al_mixer_create(freq, AL_AUDIO_32_BIT_FLOAT, chan);
			if(!mixer) {
				al_stream_destroy(stream);
				fprintf(stderr, "Error allocating mixer!\n");
				return;
			}

			if(al_voice_attach_mixer(voice, mixer) != 0) {
				al_stream_destroy(stream);
				fprintf(stderr, "Error attaching mixer to voice!\n");
				return;
			}

			/* Attach the stream to the mixer */
			if(al_mixer_attach_stream(mixer, stream) != 0) {
				al_stream_destroy(stream);
				fprintf(stderr, "Error attaching stream to mixer!\n");
				return;
			}

			display_mixer_info(mixer);

			while(!feof(f))
			{
				unsigned long waiting = 0;
				void *buffer;

				al_stream_get_long(stream, AL_AUDIO_USED_FRAGMENTS, &waiting);
				if(!waiting)
				{
					usleep(1000);
					continue;
				}
				while(waiting-- > 0)
				{
					al_stream_get_ptr(stream, AL_AUDIO_BUFFER, &buffer);
					fread(buffer, 1, bpu, f);
					al_stream_set_ptr(stream, AL_AUDIO_BUFFER, buffer);
				}
			}

			usleep(999999);
			al_stream_destroy(stream);
			stream = NULL;
			al_mixer_destroy(mixer);
			mixer = NULL;
			al_voice_destroy(voice);
			voice = NULL;

			fclose(f);
			f = NULL;
			return;
		}
	}
}


struct WAVFILE* allegro_load_wav(const char *filename) {
	FILE *f;
	struct WAVFILE *wf;

	f = fopen(filename, "rb");
	if(!f)
	{
		fprintf(stderr, "Could not open %s\n", filename);
		return NULL;
	}

	wf = malloc(sizeof(struct WAVFILE));
	wf->len = 0;
	wf->freq = 0;
	wf->bits = 0;
	wf->channels = 0;
	wf->buffer = NULL;

	/* check RIFF header */
	if(fgetc(f)!='R' || fgetc(f)!='I' || fgetc(f)!='F' || fgetc(f)!='F')
	{
		fprintf(stderr, "Bad RIFF header!\n");
		free(wf);
		return NULL;
	}
	fgetc(f); fgetc(f);
	fgetc(f); fgetc(f);
//	fseek(f, 4, SEEK_CUR);

	if(fgetc(f)!='W' || fgetc(f)!='A' || fgetc(f)!='V' || fgetc(f)!='E')
	{
		fprintf(stderr, "Bad WAVE header!\n");
		free(wf);
		return NULL;
	}

	while(1)
	{
		char tag[4];
		int length;

		if(fread(tag, 1, 4, f) != 4)
		{
			fprintf(stderr, "EOF looking for wave data!\n");
			free(wf);
			return NULL;
		}

		/* read chunk length */
		length  = fgetc(f);
		length |= fgetc(f)<<8;
		length |= fgetc(f)<<16;
		length |= fgetc(f)<<24;

		if(memcmp(tag, "fmt ", 4) == 0)
		{
			short i = fgetc(f);
			i |= fgetc(f)<<8;
			length -= 2;
			/* should be 1 for PCM data */
			if (i != 1)
			{
				fprintf(stderr, "Not PCM data!\n");
				free(wf);
				return NULL;
			}

			wf->channels  = fgetc(f);
			wf->channels |= fgetc(f)<<8;
			length -= 2;

			/* sample frequency */
			wf->freq  = fgetc(f);
			wf->freq |= fgetc(f)<<8;
			wf->freq |= fgetc(f)<<16;
			wf->freq |= fgetc(f)<<24;
			length -= 4;

			/* skip six bytes */
			fgetc(f); fgetc(f);
			fgetc(f); fgetc(f);
			fgetc(f); fgetc(f);
			length -= 6;

			wf->bits  = fgetc(f);
			wf->bits |= fgetc(f)<<8;
			length -= 2;

			fprintf(stderr, "Opened file: %d bits, %d hz, %d channel\n",
			        wf->bits, wf->freq, wf->channels);
		}
		else if (memcmp(tag, "data", 4) == 0)
		{
			if(wf->channels == 0 || wf->bits == 0) {
				free(wf);
				return NULL;
			}

			// allocate enough space even if length is odd for some reason
			wf->len  = (length + (wf->channels-1)) / wf->channels;
			wf->len /= (wf->bits+7)/8;

			free(wf->buffer);
			wf->buffer = malloc(wf->len * wf->channels * ((wf->bits+7)/8));
			if(!wf->buffer)
			{
				fprintf(stderr, "Out of memory (%d bytes)!\n", wf->len * wf->channels * (wf->bits+7) / 8);
				free(wf);
				return NULL;
			}

			if(wf->bits == 16)
			{
				int i;
				length /= 2;
				for(i = 0;i < length;++i)
				{
					((int16_t*)wf->buffer)[i]  = fgetc(f);
					((int16_t*)wf->buffer)[i] |= fgetc(f)<<8;
				}
				length *= 2;
			}
			else
				fread(wf->buffer, 1, length, f);

			break;
		}

		while(length > 0) {             /* skip the remainder of the chunk */
			fgetc(f);
			--length;
		}
	}
	fclose(f);

	return wf;
}


AL_SAMPLE* allegro_load_sample(const char *filename) {
	struct WAVFILE *wf;
	AL_SAMPLE *sample;

	wf = allegro_load_wav(filename);
	if (!wf)
		return NULL;

	sample = al_sample_create(wf->buffer, wf->len, wf->freq,
                           (wf->bits==8)?AL_AUDIO_8_BIT_UINT:AL_AUDIO_16_BIT_INT,
                           (wf->channels==1)?AL_AUDIO_1_CH:AL_AUDIO_2_CH);
	if (!sample) {
		free(wf->buffer);
		free(wf);
		return NULL;
	}

	al_sample_set_bool(sample, AL_AUDIO_ORPHAN_BUFFER, 1);
	free(wf);
	return sample;
}
