/* internal-only header */

#include "sound.h"
#include "mutex.h"

#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>



struct AL_AUDIO_DRIVER {
   const char *specifier;

   int  (*open)();
   void (*close)();

   int (*get_bool)(AL_AUDIO_ENUM, bool*);
   int (*get_enum)(AL_AUDIO_ENUM, AL_AUDIO_ENUM*);
   int (*get_long)(AL_AUDIO_ENUM, unsigned long*);
   int (*get_ptr)(AL_AUDIO_ENUM, const void**);

   int (*set_bool)(AL_AUDIO_ENUM, bool);
   int (*set_enum)(AL_AUDIO_ENUM, AL_AUDIO_ENUM);
   int (*set_long)(AL_AUDIO_ENUM, unsigned long);
   int (*set_ptr)(AL_AUDIO_ENUM, const void*);

   int  (*allocate_voice)(AL_VOICE*);
   void (*deallocate_voice)(AL_VOICE*);

   int (*load_voice)(AL_VOICE*, const void*);
   void (*unload_voice)(AL_VOICE*);

   int  (*start_voice)(AL_VOICE*);
   void (*stop_voice)(AL_VOICE*);

   bool (*voice_is_playing)(const AL_VOICE*);

   unsigned long (*get_voice_position)(const AL_VOICE*);
   int (*set_voice_position)(AL_VOICE*, unsigned long);
};

const void *_al_voice_update(AL_VOICE *voice, unsigned long samples);

/* A voice structure that you'd attach a mixer or sample to. Ideally there
   would be one AL_VOICE per system/hardware voice */
struct AL_VOICE {
   AL_AUDIO_ENUM depth;
   AL_AUDIO_ENUM chan_conf;
   AL_AUDIO_ENUM settings;

   unsigned long frequency;

   /* If set to non-0, they must be honored by the driver */
   size_t buffer_size;
   size_t num_buffers;

   /* This may be an AL_SAMPLE or AL_MIXER object */
   AL_SAMPLE *stream;
   /* True for voices with an attached mixer */
   bool streaming;

   _AL_MUTEX mutex;

   AL_AUDIO_DRIVER *driver;

   /* Extra data for use by the driver */
   void *extra;
};


typedef void (*stream_reader)(void*,void**,unsigned long,AL_AUDIO_ENUM,size_t);

/* The sample struct */
struct AL_SAMPLE {
   volatile bool playing;
   volatile bool is_stream;

   AL_AUDIO_ENUM loop;
   AL_AUDIO_ENUM depth;
   AL_AUDIO_ENUM chan_conf;
   unsigned long frequency;
   float speed;

   union {
      float    *f32;
      uint32_t *u24;
      int32_t  *s24;
      uint16_t *u16;
      int16_t  *s16;
      uint8_t  *u8;
      int8_t   *s8;
      void *ptr;
   } buffer;
   bool orphan_buffer;
   unsigned long pos, len;
   unsigned long loop_start, loop_end;
   long step;

   /* Used to convert from this format to the attached mixer's */
   float *matrix;

   /* Reads sample data into the provided buffer, using the specified format,
      converting as necessary */
   stream_reader read;

   /* The mutex is shared with the parent object. It is NULL if it is not
      directly or indirectly attached to a voice. */
   _AL_MUTEX *mutex;

   union {
      AL_MIXER *mixer;
      AL_VOICE *voice;
      void *ptr;
   } parent;
   bool parent_is_voice;
};


struct AL_STREAM {
   AL_SAMPLE spl;

   size_t buf_count;

   void *main_buffer;

   void **pending_bufs;
   void **used_bufs;
};


typedef void (*pp_callback)(void*,unsigned long,void*);

/* Slight evilness. AL_MIXER is derived from AL_SAMPLE. Certain internal
   functions and pointers may take either object type, and such things are
   explicitly noted. This is never exposed to the user, though.
   The sample object's read method will be set to a different function
   that will call the read method of all attached streams (which may be a
   sample, or another mixer) */
struct AL_MIXER {
   AL_SAMPLE ss;

   AL_AUDIO_ENUM quality;

   pp_callback post_process;
   void *cb_ptr;

   AL_SAMPLE **streams;
};
