/* user-included header */
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>

#include "alunix.h"

typedef enum AL_AUDIO_ENUM {
   /* Sample depth and type, and signedness. Mixers only use 32-bit signed
      float (-1..+1). The unsigned value is a bit-flag applied to the depth
      value */
   AL_AUDIO_8_BIT_INT    = 0x00,
   AL_AUDIO_16_BIT_INT   = 0x01,
   AL_AUDIO_24_BIT_INT   = 0x02,
   AL_AUDIO_32_BIT_FLOAT = 0x03,

   AL_AUDIO_UNSIGNED   = 0x08,

   /* For convenience */
   AL_AUDIO_8_BIT_UINT   = AL_AUDIO_8_BIT_INT|AL_AUDIO_UNSIGNED,
   AL_AUDIO_16_BIT_UINT  = AL_AUDIO_16_BIT_INT|AL_AUDIO_UNSIGNED,
   AL_AUDIO_24_BIT_UINT  = AL_AUDIO_24_BIT_INT|AL_AUDIO_UNSIGNED,

   /* Speaker configuration (mono, stereo, 2.1, 3, etc). With regards to
      behavior, most of this code makes no distinction between, say, 4.1 and
      5 speaker setups.. they both have 5 "channels". However, users would
      like the distinction, and later when the higher-level stuff is added,
      the differences will become more important. (v>>4)+(v&0xF) should yield
      the total channel count */
   AL_AUDIO_1_CH   = 0x10,
   AL_AUDIO_2_CH   = 0x20,
   AL_AUDIO_2_1_CH = 0x21,
   AL_AUDIO_3_CH   = 0x30,
   AL_AUDIO_3_1_CH = 0x31,
   AL_AUDIO_4_CH   = 0x40,
   AL_AUDIO_4_1_CH = 0x41,
   AL_AUDIO_5_CH   = 0x50,
   AL_AUDIO_5_1_CH = 0x51,
   AL_AUDIO_6_CH   = 0x60,
   AL_AUDIO_6_1_CH = 0x61,
   AL_AUDIO_7_CH   = 0x70,

   /* Sample looping mode */
   AL_AUDIO_PLAY_ONCE = 0x100,
   AL_AUDIO_ONE_DIR   = 0x101,
   AL_AUDIO_BI_DIR    = 0x102,

   /* Mixer quality */
   AL_AUDIO_POINT  = 0x110,
   AL_AUDIO_LINEAR = 0x111,

   /* Bit-field flags for voice settings. This says what kind of voice you
      want/have, either a hardware or software one, and whether the voice
      settings were required */
   AL_AUDIO_HARDWARE = 0x121,

   AL_AUDIO_REQUIRE  = 0x122,

   /* Flags to pass to the various al_*_get_* and al_*_set_* functions. Not
      all types will apply to all functions. */
   AL_AUDIO_DEPTH     = 0x200,
   AL_AUDIO_CHANNELS  = 0x201,
   AL_AUDIO_FREQUENCY = 0x202,

   AL_AUDIO_PLAYING  = 0x203,
   AL_AUDIO_ATTACHED = 0x204,

   AL_AUDIO_LENGTH = 0x205,
   AL_AUDIO_BUFFER = 0x206,

   AL_AUDIO_LOOPMODE      = 0x207,
   AL_AUDIO_SPEED         = 0x208,
   AL_AUDIO_POSITION      = 0x209,
   AL_AUDIO_ORPHAN_BUFFER = 0x20A,

   AL_AUDIO_FRAGMENTS      = 0x20B,
   AL_AUDIO_USED_FRAGMENTS = 0x20C,

   AL_AUDIO_QUALITY = 0x20D,

   AL_AUDIO_SETTINGS = 0x20E,

   AL_AUDIO_DEVICE = 0x20F,

   /* Used to define driver-specific option values, not to collide with standard values. */
   AL_AUDIO_USER_START = 0x10000
} AL_AUDIO_ENUM;

/* struct AL_SAMPLE:

An AL_SAMPLE object stores the data necessary for playing pre-defined digital audio. It holds information pertaining to data length, the looping mode, loop start/end points, playing position, playing frequency, channel configuration, etc. As such, you cannot have an AL_SAMPLE object playing multiple times simultaniously. The object holds a user-specified PCM data buffer, of the format the object is created with. Multiple objects may be created using the same data buffer.

To be played, an AL_SAMPLE object must be attached to an AL_VOICE object, or to an AL_MIXER object which is itself attached to an AL_VOICE object (or to another AL_MIXER object which is attached to an AL_VOICE object, etc).

An AL_SAMPLE object uses the following fields:

* AL_AUDIO_DEPTH (enum) - Gets the bit depth format the object was created with. This may not be changed after the object is created.

* AL_AUDIO_CHANNELS (enum) - Gets the channel configuration the object was created with. This may not be changed after the object is created.

* AL_AUDIO_FREQUENCY (long) - Gets the frequency (in hz) the object was created with. This may not be changed after the object is created. To change playback speed, see AL_AUDIO_SPEED.

* AL_AUDIO_PLAYING (bool) - Gets or sets the object's playing status. By default, it is stopped. Note that simply setting it true does not cause the object to play. It must also be attached to a voice, directly or indirectly (eg. sample->voice, sample->mixer->voice, sample->mixer->...->voice).

* AL_AUDIO_ATTACHED (bool) - Gets the object's attachment status (true if it is attached to a something, false if not). Setting this to false detaches the object from whatever it is attached to. You may not directly set this to true.

* AL_AUDIO_LENGTH (long) - Gets or sets the length of the object's data buffer, in samples-per-channel. When changing the length, you must make sure the current buffer is large enough. You may not change the length while the object is set to play.

* AL_AUDIO_BUFFER (ptr) - Gets or sets the object's data buffer. You may not get or set this if the object is set to play, and you may not get this if the buffer was previously orphaned (see AL_AUDIO_ORPHAN_BUFFER).

* AL_AUDIO_LOOPMODE (enum) - Gets or sets the object's looping mode. Setting this may fail if the object is attached to a voice and the audio driver does not support the requested looping mode.

* AL_AUDIO_SPEED (float) - Gets or sets the object's playing speed. Negative values will cause the object to play backwards. If the value is set too close to 0, this will fail to set.

* AL_AUDIO_POSITION (long) - Gets or sets the object's playing position. The value is in samples-per-channel.

* AL_AUDIO_ORPHAN_BUFFER (bool) - Setting this flag to true will cause the object's data buffer to be free()'d automatically when either the sample is destroyed, or the buffer is changed (setting an AL_AUDIO_BUFFER value). If you set another AL_AUDIO_BUFFER value, this will revert back to false. You may not directly set this to false. You must not orphan a buffer that is referenced outside of the object.

*/
typedef struct AL_SAMPLE AL_SAMPLE;

/* AL_STREAM:

An AL_STREAM object is used to stream generated audio to the sound device, in real-time. As with AL_SAMPLE objects, they store information necessary for playback, so you may not play one multiple times simultaniously. They also need to be attached to an AL_VOICE object, or to an AL_MIXER object which, eventually, reaches an AL_VOICE object.

While playing, you must periodicly supply new buffer data by first checking AL_AUDIO_USED_FRAGMENTS, then refilling the buffers via AL_AUDIO_BUFFER. If you're late with supplying new data, the object will be silenced until new data is provided.

AL_STREAM objects use the following fields:

* AL_AUDIO_DEPTH (enum) - Same as AL_SAMPLE

* AL_AUDIO_CHANNELS (enum) - Same as AL_SAMPLE

* AL_AUDIO_FREQUENCY (enum) - Same as AL_SAMPLE

* AL_AUDIO_ATTACHED (bool) - Same as AL_SAMPLE

* AL_AUDIO_PLAYING (bool) - Same as AL_SAMPLE, with the exception that AL_STREAM objects are set to play by default.

* AL_AUDIO_SPEED (float) - Same as AL_SAMPLE, with the added caveat that negative values aren't allowed.

* AL_AUDIO_LENGTH (long) - This gets the length, in samples-per-channel, of the individual buffer fragments. You may not set this after the object is created.

* AL_AUDIO_BUFFER (ptr) - This gets the next buffer fragment that needs to be filled. After the buffer is filled, this field must be set to the same pointer value to let the object know the new data is ready.

* AL_AUDIO_FRAGMENTS (long) - This gets the total number of buffer fragments the object was created with. You may not set this after the object is created.

* AL_AUDIO_USED_FRAGMENTS (long) - This gets the number of buffer fragments that are waiting to be refilled. This value is decreased when AL_AUDIO_BUFFER is used to retrieve a waiting buffer fragment. You may not set this value.

*/
typedef struct AL_STREAM AL_STREAM;

/* struct AL_MIXER:

*/
typedef struct AL_MIXER AL_MIXER;

/* struct AL_VOICE:

*/
typedef struct AL_VOICE AL_VOICE;

/* struct AL_AUDIO_DRIVER:

*/
typedef struct AL_AUDIO_DRIVER AL_AUDIO_DRIVER;

/* Sample functions (more detailed explanations below) */
AL_SAMPLE *al_sample_create(void *buf, unsigned long samples, unsigned long freq, AL_AUDIO_ENUM depth, AL_AUDIO_ENUM chan_conf);
void al_sample_destroy(AL_SAMPLE *spl);

int al_sample_get_long(const AL_SAMPLE *spl, AL_AUDIO_ENUM setting, unsigned long *val);
int al_sample_get_float(const AL_SAMPLE *spl, AL_AUDIO_ENUM setting, float *val);
int al_sample_get_enum(const AL_SAMPLE *spl, AL_AUDIO_ENUM setting, AL_AUDIO_ENUM *val);
int al_sample_get_bool(const AL_SAMPLE *spl, AL_AUDIO_ENUM setting, bool *val);
int al_sample_get_ptr(const AL_SAMPLE *spl, AL_AUDIO_ENUM setting, void **ptr);

int al_sample_set_long(AL_SAMPLE *spl, AL_AUDIO_ENUM setting, unsigned long val);
int al_sample_set_float(AL_SAMPLE *spl, AL_AUDIO_ENUM setting, float val);
int al_sample_set_enum(AL_SAMPLE *spl, AL_AUDIO_ENUM setting, AL_AUDIO_ENUM val);
int al_sample_set_bool(AL_SAMPLE *spl, AL_AUDIO_ENUM setting, bool val);
int al_sample_set_ptr(AL_SAMPLE *spl, AL_AUDIO_ENUM setting, void *ptr);

void al_sample_play(AL_SAMPLE *spl);
void al_sample_stop(AL_SAMPLE *spl);

/* Stream functions */
AL_STREAM *al_stream_create(size_t buffer_count, unsigned long samples, unsigned long freq, AL_AUDIO_ENUM depth, AL_AUDIO_ENUM chan_conf);
void al_stream_destroy(AL_STREAM *stream);

int al_stream_get_long(const AL_STREAM *stream, AL_AUDIO_ENUM setting, unsigned long *val);
int al_stream_get_float(const AL_STREAM *stream, AL_AUDIO_ENUM setting, float *val);
int al_stream_get_enum(const AL_STREAM *stream, AL_AUDIO_ENUM setting, AL_AUDIO_ENUM *val);
int al_stream_get_bool(const AL_STREAM *stream, AL_AUDIO_ENUM setting, bool *val);
int al_stream_get_ptr(const AL_STREAM *spl, AL_AUDIO_ENUM setting, void **ptr);

int al_stream_set_long(AL_STREAM *stream, AL_AUDIO_ENUM setting, unsigned long val);
int al_stream_set_float(AL_STREAM *stream, AL_AUDIO_ENUM setting, float val);
int al_stream_set_enum(AL_STREAM *stream, AL_AUDIO_ENUM setting, AL_AUDIO_ENUM val);
int al_stream_set_bool(AL_STREAM *stream, AL_AUDIO_ENUM setting, bool val);
int al_stream_set_ptr(AL_STREAM *spl, AL_AUDIO_ENUM setting, void *ptr);

/* Mixer functions */
AL_MIXER *al_mixer_create(unsigned long freq, AL_AUDIO_ENUM depth, AL_AUDIO_ENUM chan_conf);
void al_mixer_destroy(AL_MIXER *mixer);

int al_mixer_attach_sample(AL_MIXER *mixer, AL_SAMPLE *stream);
int al_mixer_attach_stream(AL_MIXER *mixer, AL_STREAM *stream);
int al_mixer_attach_mixer(AL_MIXER *mixer, AL_MIXER *stream);

int al_mixer_set_postprocess_callback(AL_MIXER *mixer, void (*cb)(void *buf, unsigned long samples, void *data), void *data);

int al_mixer_get_long(const AL_MIXER *mixer, AL_AUDIO_ENUM setting, unsigned long *val);
int al_mixer_get_enum(const AL_MIXER *mixer, AL_AUDIO_ENUM setting, AL_AUDIO_ENUM *val);
int al_mixer_get_bool(const AL_MIXER *mixer, AL_AUDIO_ENUM setting, bool *val);

int al_mixer_set_long(AL_MIXER *mixer, AL_AUDIO_ENUM setting, unsigned long val);
int al_mixer_set_enum(AL_MIXER *mixer, AL_AUDIO_ENUM setting, AL_AUDIO_ENUM val);
int al_mixer_set_bool(AL_MIXER *mixer, AL_AUDIO_ENUM setting, bool val);

/* Voice functions */
AL_VOICE *al_voice_create(AL_AUDIO_DRIVER *driver, unsigned long freq, AL_AUDIO_ENUM depth, AL_AUDIO_ENUM chan_conf, AL_AUDIO_ENUM settings);
void al_voice_destroy(AL_VOICE *voice);

int al_voice_attach_sample(AL_VOICE *voice, AL_SAMPLE *stream);
int al_voice_attach_mixer(AL_VOICE *voice, AL_MIXER *mixer);

void al_voice_detach(AL_VOICE *voice);

int al_voice_get_long(const AL_VOICE *voice, AL_AUDIO_ENUM setting, unsigned long *val);
int al_voice_get_enum(const AL_VOICE *voice, AL_AUDIO_ENUM setting, AL_AUDIO_ENUM *val);
int al_voice_get_bool(const AL_VOICE *voice, AL_AUDIO_ENUM setting, bool *val);

int al_voice_set_long(AL_VOICE *voice, AL_AUDIO_ENUM setting, unsigned long val);
int al_voice_set_enum(AL_VOICE *voice, AL_AUDIO_ENUM setting, AL_AUDIO_ENUM val);
int al_voice_set_bool(AL_VOICE *voice, AL_AUDIO_ENUM setting, bool val);


/* Misc. audio functions */
AL_AUDIO_DRIVER *al_get_audio_driver(const char *name);
AL_AUDIO_DRIVER *al_audio_init_driver(AL_AUDIO_DRIVER *driver);
void al_audio_deinit_driver(AL_AUDIO_DRIVER *driver);

int al_audio_get_bool(AL_AUDIO_DRIVER *driver, AL_AUDIO_ENUM setting, bool *val);
int al_audio_get_enum(AL_AUDIO_DRIVER *driver, AL_AUDIO_ENUM setting, AL_AUDIO_ENUM *val);
int al_audio_get_long(AL_AUDIO_DRIVER *driver, AL_AUDIO_ENUM setting, unsigned long *val);
int al_audio_get_ptr(AL_AUDIO_DRIVER *driver, AL_AUDIO_ENUM setting, const void **val);

int al_audio_set_bool(AL_AUDIO_DRIVER *driver, AL_AUDIO_ENUM setting, bool val);
int al_audio_set_enum(AL_AUDIO_DRIVER *driver, AL_AUDIO_ENUM setting, AL_AUDIO_ENUM val);
int al_audio_set_long(AL_AUDIO_DRIVER *driver, AL_AUDIO_ENUM setting, unsigned long val);
int al_audio_set_ptr(AL_AUDIO_DRIVER *driver, AL_AUDIO_ENUM setting, const void *val);

